'''
Created on 2020-08-19

@author: wf
'''
from lodstorage.yamlablemixin import YamlAbleMixin
from lodstorage.jsonpicklemixin import JsonPickleMixin
from lodstorage.jsonable import JSONAble
from lodstorage.storageconfig import StorageConfig, StoreMode
from lodstorage.sparql import SPARQL
from lodstorage.sql import SQLDB
from lodstorage.lod import LOD
from lodstorage.jsonable import JSONAbleList
import sys
import os
import time

class EntityManager(YamlAbleMixin, JsonPickleMixin,JSONAbleList):
    '''
    generic entity manager
    '''

    def __init__(self,name,entityName,entityPluralName:str,listName:str=None,clazz=None,tableName:str=None,primaryKey:str=None,config=None,handleInvalidListTypes=False,filterInvalidListTypes=False,listSeparator='⇹',debug=False):
        '''
        Constructor
        
        Args:
            name(string): name of this eventManager
            entityName(string): entityType to be managed e.g. Country
            entityPluralName(string): plural of the the entityType e.g. Countries
            config(StorageConfig): the configuration to be used if None a default configuration will be used
            handleInvalidListTypes(bool): True if invalidListTypes should be converted or filtered
            filterInvalidListTypes(bool): True if invalidListTypes should be deleted
            listSeparator(str): the symbol to use as a list separator
            debug(boolean): override debug setting when default of config is used via config=None
        '''
        self.name=name
        self.entityName=entityName
        self.entityPluralName=entityPluralName
        if listName is None:
            listName=entityPluralName
        if tableName is None:
            tableName=entityName
        self.primaryKey=primaryKey
        if config is None:
            config=StorageConfig.getDefault()
            if debug:
                config.debug=debug
        self.config=config
        super(EntityManager, self).__init__(listName=listName,clazz=clazz,tableName=tableName,handleInvalidListTypes=handleInvalidListTypes,filterInvalidListTypes=filterInvalidListTypes)
        cacheFile=self.getCacheFile(config=config,mode=config.mode)
        self.showProgress ("Creating %smanager(%s) for %s using cache %s" % (self.entityName,config.mode,self.name,cacheFile))
        if config.mode is StoreMode.SPARQL:
            if config.endpoint is None:
                raise Exception("no endpoint set for mode sparql") 
            self.endpoint=config.endpoint   
            self.sparql=SPARQL(config.endpoint,debug=config.debug,profile=config.profile)
        elif config.mode is StoreMode.SQL:
            self.executeMany=False # may be True when issues are fixed
        self.listSeparator=listSeparator
            
    def storeMode(self):
        '''
        return my store mode
        '''
        return self.config.mode

    def showProgress(self,msg):
        ''' display a progress message 
            
            Args:
              msg(string): the message to display
        '''
        if self.config.withShowProgress:
            print (msg,flush=True)      
     
    def getCacheFile(self,config=None,mode=StoreMode.SQL):
        '''
        get the cache file for this event manager
        Args:
            config(StorageConfig): if None get the cache for my mode
            mode(StoreMode): the storeMode to use
        '''
        if config is None:
            config=self.config
        cachedir=config.getCachePath() 
        if config.cacheFile is not None:
            return config.cacheFile
        ''' get the path to the file for my cached data '''  
        if mode is StoreMode.JSON or mode is StoreMode.JSONPICKLE:  
            extension=f".{mode.name.lower()}"
            cachepath=f"{cachedir}/{self.name}-{self.listName}{extension}" 
        elif mode is StoreMode.SPARQL:
            cachepath=f"SPAQRL {self.name}:{config.endpoint}"    
        elif mode is StoreMode.SQL:
            cachepath=f"{cachedir}/{self.name}.db"
        else:
            cachepath=f"undefined cachepath for StoreMode {mode}"
        return cachepath     
    
    def removeCacheFile(self):
        '''  remove my cache file '''
        mode=self.config.mode
        if mode is StoreMode.JSON or mode is StoreMode.JSONPICKLE:
            cacheFile=self.getCacheFile(mode=mode)
            if os.path.isfile(cacheFile):
                os.remove(cacheFile)
                    
    def getSQLDB(self,cacheFile):
        '''
        get the SQL database for the given cacheFile
        
        Args:
            cacheFile(string): the file to get the SQL db from
        '''
        config=self.config
        sqldb=self.sqldb=SQLDB(cacheFile,debug=config.debug,errorDebug=config.errorDebug)
        return sqldb
    
    def initSQLDB(self,sqldb,listOfDicts=None,withCreate:bool=True,withDrop:bool=True,sampleRecordCount=-1):
        '''
        initialize my sql DB  
        
        Args:
            listOfDicts(list): the list of dicts to analyze for type information
            withDrop(boolean): true if the existing Table should be dropped
            withCreate(boolean): true if the create Table command should be executed - false if only the entityInfo should be returned
            sampleRecordCount(int): the number of records to analyze for type information
        Return:
            EntityInfo: the entity information such as CREATE Table command
        '''
        if listOfDicts is None:
            listOfDicts=JSONAble.getJsonTypeSamplesForClass(self.clazz)
        entityInfo=sqldb.createTable(listOfDicts, self.tableName, primaryKey=self.primaryKey,withCreate=withCreate,withDrop=withDrop,sampleRecordCount=sampleRecordCount) 
        return entityInfo
    
    def setNone(self,record,fields):
        '''
        make sure the given fields in the given record are set to none
        Args:
            record(dict): the record to work on
            fields(list): the list of fields to set to None 
        '''
        LOD.setNone(record,fields)
            
    def isCached(self):
        ''' check whether there is a file containing cached 
        data for me '''
        result=False
        config=self.config
        mode=self.config.mode
        if mode is StoreMode.JSON or mode is StoreMode.JSONPICKLE:
            result=os.path.isfile(self.getCacheFile(config=self.config,mode=mode))
        elif mode is StoreMode.SPARQL:
            # @FIXME - make abstract
            query=config.prefix+"""
SELECT  ?source (COUNT(?source) AS ?sourcecount)
WHERE { 
   ?event cr:Event_source ?source.
}
GROUP by ?source
"""                                 
            sourceCountList=self.sparql.queryAsListOfDicts(query)
            for sourceCount in sourceCountList:
                source=sourceCount['source'];
                recordCount=sourceCount['sourcecount']
                if source==self.name and recordCount>100:
                    result=True
        elif mode is StoreMode.SQL:
            cacheFile=self.getCacheFile(config=self.config,mode=StoreMode.SQL)
            if os.path.isfile(cacheFile):
                sqlQuery=f"SELECT COUNT(*) AS count FROM {self.tableName}" 
                try:
                    sqlDB=self.getSQLDB(cacheFile)
                    countResults=sqlDB.query(sqlQuery)
                    countResult=countResults[0]
                    count=countResult['count']
                    result=count>=0
                except Exception as ex:
                    msg=str(ex)
                    if self.debug:
                        print(msg,file=sys.stderr)
                        sys.stderr.flush()
                    # e.g. sqlite3.OperationalError: no such table: Event_crossref
                    pass      
        else:
            raise Exception("unsupported mode %s" % self.mode)            
        return result     
    
    def fromCache(self,force:bool=False,getListOfDicts=None,append=False,sampleRecordCount=-1):
        '''
        get my entries from the cache or from the callback provided
        
        Args:
            force(bool): force ignoring the cache
            getListOfDicts(callable): a function to call for getting the data 
            append(bool): True if records should be appended
            sampleRecordCount(int): the number of records to analyze for type information
            
        Returns:
            the list of Dicts and as a side effect setting self.cacheFile
        '''
        if not self.isCached() or force:
            startTime=time.time()
            self.showProgress(f"getting {self.entityPluralName} for {self.name} ...")
            if getListOfDicts is None:
                if hasattr(self, "getListOfDicts"):
                    getListOfDicts=self.getListOfDicts
                else:
                    raise  Exception("from Cache failed and no secondary cache via getListOfDicts specified")
            listOfDicts=getListOfDicts()
            duration=time.time()-startTime
            self.showProgress(f"got {len(listOfDicts)} {self.entityPluralName} in {duration:5.1f} s")   
            self.cacheFile=self.storeLoD(listOfDicts,append=append,sampleRecordCount=sampleRecordCount)
            self.setListFromLoD(listOfDicts)
        else:
            # fromStore also sets self.cacheFile
            listOfDicts=self.fromStore()
        return listOfDicts
        
    def fromStore(self,cacheFile=None,setList:bool=True)->list:
        '''
        restore me from the store
        Args:
            cacheFile(String): the cacheFile to use if None use the pre configured cachefile
            setList(bool): if True set my list with the data from the cache file
            
        Returns:
            list: list of dicts or JSON entitymanager
        '''
        startTime=time.time()
        if cacheFile is None:
            cacheFile=self.getCacheFile(config=self.config,mode=self.config.mode)
        self.cacheFile=cacheFile
        self.showProgress("reading %s for %s from cache %s" % (self.entityPluralName,self.name,cacheFile))
        mode=self.config.mode
        if mode is StoreMode.JSONPICKLE:   
            JSONem=JsonPickleMixin.readJsonPickle(cacheFile)
            if self.clazz is not None:
                listOfDicts=JSONem.getLoD()
            else:
                listOfDicts=JSONem.getList()
        elif mode is StoreMode.JSON:
            listOfDicts=self.readLodFromJsonFile(cacheFile)
            pass
        elif mode is StoreMode.SPARQL:
            # @FIXME make abstract
            eventQuery="""
PREFIX cr: <http://cr.bitplan.com/>
SELECT ?eventId ?acronym ?series ?title ?year ?country ?city ?startDate ?endDate ?url ?source WHERE { 
   OPTIONAL { ?event cr:Event_eventId ?eventId. }
   OPTIONAL { ?event cr:Event_acronym ?acronym. }
   OPTIONAL { ?event cr:Event_series ?series. }
   OPTIONAL { ?event cr:Event_title ?title. }
   OPTIONAL { ?event cr:Event_year ?year.  }
   OPTIONAL { ?event cr:Event_country ?country. }
   OPTIONAL { ?event cr:Event_city ?city. }
   OPTIONAL { ?event cr:Event_startDate ?startDate. }
   OPTIONAL { ?event cr:Event_endDate ?endDate. }
   OPTIONAL { ?event cr:Event_url ?url. }
   ?event cr:Event_source ?source FILTER(?source='%s').
}
""" % self.name        
            listOfDicts=self.sparql.queryAsListOfDicts(eventQuery)
        elif mode is StoreMode.SQL:
            sqlQuery="SELECT * FROM %s" % self.tableName
            sqlDB=self.getSQLDB(cacheFile)
            listOfDicts=sqlDB.query(sqlQuery)
            sqlDB.close()
            pass
        else:
            raise Exception("unsupported store mode %s" % self.mode)
          
        self.showProgress("read %d %s from %s in %5.1f s" % (len(listOfDicts),self.entityPluralName,self.name,time.time()-startTime))
        if setList:
            self.setListFromLoD(listOfDicts)
        return listOfDicts
    
    def getLoD(self):
        """
        Return the LoD of the entities in the list
        
        Return:
            list: a list of Dicts
            
        """
        lod= []
        for entity in self.getList():
            # TODO - optionally filter by samples
            lod.append(entity.__dict__)
        return lod
    
    def store(self,limit=10000000,batchSize=250,append=False,fixNone=True,sampleRecordCount=-1)->str:
        '''
        store my list of dicts
        
        Args:
            limit(int): maximum number of records to store per batch
            batchSize(int): size of batch for storing
            append(bool): True if records should be appended
            fixNone(bool): if True make sure the dicts are filled with None references for each record
            sampleRecordCount(int): the number of records to analyze for type information
        
        Return:
            str: The cachefile being used
        '''
        lod=self.getLoD()
        return self.storeLoD(lod,limit=limit,batchSize=batchSize,append=append,fixNone=fixNone,sampleRecordCount=sampleRecordCount)
        
    def storeLoD(self,listOfDicts,limit=10000000,batchSize=250,cacheFile=None,append=False,fixNone=True,sampleRecordCount=1)->str:
        ''' 
        store my entities 
        
        Args:
            listOfDicts(list): the list of dicts to store
            limit(int): maximum number of records to store
            batchSize(int): size of batch for storing
            cacheFile(string): the name of the storage e.g path to JSON or sqlite3 file
            append(bool): True if records should be appended
            fixNone(bool): if True make sure the dicts are filled with None references for each record
            sampleRecordCount(int): the number of records to analyze for type information
            
        Return:
            str: The cachefile being used
        '''
        config=self.config
        mode=config.mode
        if self.handleInvalidListTypes:
            LOD.handleListTypes(lod=listOfDicts,doFilter=self.filterInvalidListTypes,separator=self.listSeparator)
        if mode is StoreMode.JSON or mode is StoreMode.JSONPICKLE:    
            if cacheFile is None:
                cacheFile=self.getCacheFile(config=self.config,mode=mode)
            self.showProgress (f"storing {len(listOfDicts)} {self.entityPluralName} for {self.name} to cache {cacheFile}")
            if mode is StoreMode.JSONPICKLE:
                self.writeJsonPickle(cacheFile)
            if mode is StoreMode.JSON:
                self.storeToJsonFile(cacheFile)
                pass
        elif mode is StoreMode.SPARQL:
            startTime=time.time()
            msg=f"storing {len(listOfDicts)} {self.entityPluralName} to {self.config.mode} ({self.config.endpoint})"
            self.showProgress (msg) 
            # @ FIXME make abstract /configurable
            entityType="cr:Event"
            prefixes=self.config.prefix
            self.sparql.insertListOfDicts(listOfDicts, entityType, self.primaryKey, prefixes,limit=limit,batchSize=batchSize)
            self.showProgress ("store for %s done after %5.1f secs" % (self.name,time.time()-startTime))
        elif mode is StoreMode.SQL:
            startTime=time.time()
            if cacheFile is None:
                cacheFile=self.getCacheFile(config=self.config,mode=self.config.mode)
            sqldb=self.getSQLDB(cacheFile)
            self.showProgress ("storing %d %s for %s to %s:%s" % (len(listOfDicts),self.entityPluralName,self.name,config.mode,cacheFile)) 
            if append:
                withDrop=False
                withCreate=False
            else:
                withDrop=True
                withCreate=True
            entityInfo=self.initSQLDB(sqldb,listOfDicts,withCreate=withCreate,withDrop=withDrop,sampleRecordCount=sampleRecordCount)
            self.sqldb.store(listOfDicts, entityInfo,executeMany=self.executeMany,fixNone=fixNone)
            self.showProgress ("store for %s done after %5.1f secs" % (self.name,time.time()-startTime))
        else:
            raise Exception("unsupported store mode %s" % self.mode)  
        return cacheFile


'''
Created on 2020-08-21

@author: wf
'''
# redudant copy
# original is at
# https://github.com/WolfgangFahl/py-3rdparty-mediawiki/blob/master/wikibot/mwTable.py

class MediaWikiTable(object):
    '''
    helper for https://www.mediawiki.org/wiki/Help:Tables
    '''


    def __init__(self,wikiTable=True,colFormats=None,sortable=True,withNewLines=False):
        '''
        Constructor
        '''
        self.colFormats=colFormats
        cssDelim=""
        if wikiTable:
            cWikiTable="wikitable"
            cssDelim=" "
        else: 
            cWikiTable=""
        if sortable:
            cSortable="sortable"
        else:
            cSortable=""           
            
        self.start='{|class="%s%s%s"\n' % (cWikiTable,cssDelim,cSortable)
        self.header=None
        self.content=""
        self.end="\n|}\n"
        self.withNewLines=withNewLines
        pass
    
    def addHeader(self,record):
        '''
        add the given record as a "sample" header
        '''
        if self.withNewLines:
            headerStart="|+"
            firstColDelim="\n!"
            colDelim=firstColDelim
        else:
            headerStart="|+\n"
            firstColDelim="!"
            colDelim="!!"
        self.header=headerStart
        first=True
        for key in record.keys():
            if first:
                delim=firstColDelim
                first=False
            else:
                delim=colDelim
            self.header+="%s%s" % (delim,key)
      
    def addRow4Dict(self,record):
        if self.header is None:
            self.addHeader(record)
        if self.withNewLines:
            rowStart="\n|-"
            colDelim="\n|"
        else:
            rowStart="\n|-\n"
            colDelim="||"
        self.content+=rowStart    
        for key in record.keys():
            value=record[key]
            if self.colFormats is not None and key in self.colFormats:
                colFormat=self.colFormats[key]
            else:
                colFormat="%s"    
            self.content+=("%s"+colFormat) % (colDelim,value)
     
    def fromListOfDicts(self,listOfDicts):
        for record in listOfDicts:
            self.addRow4Dict(record)
        pass
    
    def noneReplace(self,value):
        return "" if value is None else value;
    
    def asWikiMarkup(self):
        '''
        convert me to MediaWiki markup
        
        Returns:
            string: the MediWiki Markup for this table
        '''
        markup=self.noneReplace(self.start)+ \
            self.noneReplace(self.header)+ \
            self.noneReplace(self.content)+ \
            self.noneReplace(self.end)
        return markup


#!/usr/bin/python3
# part of https://github.com/WolfgangFahl/ProceedingsTitleParser

# Yaml persistence
import yaml
import io
import os


class YamlAbleMixin(object):
    """ allow reading and writing derived objects from a yaml file"""
    debug = False

    # read me from a yaml file
    @staticmethod
    def readYaml(name):
        yamlFile = name
        if not yamlFile.endswith(".yaml"):
            yamlFile = yamlFile + ".yaml"
        # is there a yamlFile for the given name
        if os.path.isfile(yamlFile):
            with io.open(yamlFile, 'r') as stream:
                if YamlAbleMixin.debug:
                    print("reading %s" % (yamlFile))
                result = yaml.load(stream, Loader=yaml.Loader)
                if (YamlAbleMixin.debug):
                    print (result)
                return result
        else:
            return None

    # write me to my yaml file
    def writeYaml(self, name):
        yamlFile = name
        if not yamlFile.endswith(".yaml"):
            yamlFile = yamlFile + ".yaml"
        with io.open(yamlFile, 'w', encoding='utf-8') as stream:
            yaml.dump(self, stream)
            if YamlAbleMixin.debug:
                print (yaml.dump(self))



'''
Created on 2022-06-20

see  
    https://github.com/tyleradams/json-toolkit
    https://stackoverflow.com/questions/36021526/converting-an-array-dict-to-xml-in-python

@author: tyleradams
@author: wf
'''
from dicttoxml2 import dicttoxml
from xml.dom.minidom import parseString


class Lod2Xml():
    '''
    convert a list of dicts to XML
    '''
    
    def __init__(self,lod,root:str="root",node_name:callable=(lambda x:'node')):
        '''
        construct me with the given list of dicts
        
        Args:
            lod(list): the list of dicts to convert to XML
            root(str): the name of the root nod
            item_name(func): the function to use to calculate node names
        '''
        self.lod=lod
        self.root=root
        self.item_name=node_name
        
        
    def asXml(self,pretty:bool=True):
        ''' 
        convert result to XML
        
        Args:
            pretty(bool): if True pretty print the result
            
        '''
        xml=dicttoxml(self.lod, custom_root=self.root,item_func=self.item_name, attr_type=False)
        if pretty:
            dom=parseString(xml)
            prettyXml=dom.toprettyxml()
        else:
            prettyXml=xml    
        return prettyXml
        


'''
Created on 2020-08-22

@author: wf
'''

from enum import Enum 
import os
import yaml
from tabulate import tabulate
import urllib
import copy
#from wikibot.mwTable import MediaWikiTable
# redundant copy in this library to avoid dependency issues
# original is at
from lodstorage.jsonable import JSONAble
from lodstorage.mwTable import MediaWikiTable
from pylatexenc.latexencode import unicode_to_latex
import re
import sys
from pathlib import Path
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters.html import HtmlFormatter
from pygments.formatters.latex import LatexFormatter


class Format(Enum):
    '''
    the supported formats for the results to be delivered
    '''
    csv = 'csv'
    json = 'json'
    xml = 'xml'
    tsv = 'tsv'
    latex = 'latex'
    mediawiki= 'mediawiki'
    github = 'github'
 
    def __str__(self):
        return self.value
    
class YamlPath:
    @staticmethod
    def getPaths(yamlFileName:str,yamlPath:str=None):
        if yamlPath is None:
            yamlPath = f"{os.path.dirname(__file__)}/../sampledata/{yamlFileName}"
        yamlPaths=[yamlPath]
        home = str(Path.home())
        # additional yamls from users yaml configuration
        homepath = f"{home}/.pylodstorage/{yamlFileName}"
        if os.path.isfile(homepath):
            yamlPaths.append(homepath)
        return yamlPaths

class ValueFormatter():
    '''
    a value Formatter
    '''
    home = str(Path.home())
    # additional endpoints from users endpoint configuration
    formatsPath=f"{os.path.dirname(__file__)}/../sampledata/formats.yaml"
    valueFormats=None
    
    def __init__(self,name:str,formatString:str,regexps:list=None,):
        '''
        constructor
        
        Args:
            fstring(str): the format String to use
            regexps(list): the regular expressions to apply
        '''
        self.name=name
        self.regexps=regexps
        self.formatString=formatString
    
    @classmethod    
    def fromDict(cls,name:str,record:dict):
        '''
        create a ValueFormatter from the given dict
        '''
        if "regexps" in  record:
            regexps=record["regexps"]
        else:
            regexps=[]
        vf=ValueFormatter(name=name,formatString=record["format"],regexps=regexps)
        return vf
     
    @classmethod    
    def getFormats(cls,formatsPath:str=None)->dict:
        '''
        get the available ValueFormatters
        
        Args:
            formatsPath(str): the path to the yaml file to read the format specs from
        Returns:
            dict: a map for ValueFormatters by formatter Name
        '''
        if cls.valueFormats is None:
            valueFormats={}
            formatPaths=YamlPath.getPaths("formats.yaml",formatsPath)
            for formatPath in formatPaths:
                with open(formatPath, 'r') as stream:
                    valueFormatRecords = yaml.safe_load(stream)
                    for valueFormatKey,valueFormatRecord in valueFormatRecords.items():
                        valueFormats[valueFormatKey]=ValueFormatter.fromDict(name=valueFormatKey,record=valueFormatRecord)           
            cls.valueFormats=valueFormats
        return cls.valueFormats
        
    def applyFormat(self,record,key,resultFormat:Format):
        '''
        apply the given format to the given record
        
        Args:
            record(dict): the record to handle
            key(str): the property key 
            resultFormat(str): the resultFormat Style to apply
        '''
        if key in record:
            value=record[key]
            if value is not None and isinstance(value,str):
                # if there are no regular expressions specified always format
                doformat=len(self.regexps)==0
                for regexp in self.regexps:
                    try:
                        vmatch=re.match(regexp,value)
                        if (vmatch):
                            # we found a match and will format it if the value is not none
                            doformat=True
                            value=vmatch.group("value")
                    except Exception as ex:
                        print(f"ValueFormatter: {self.name}\nInvalid regular expression:{regexp}\n{str(ex)}",file=sys.stderr)
                if value is not None and doformat:
                    link=self.formatString.format(value=value)
                    newValue=None
                    if resultFormat=="github":
                        newValue=f"[{value}]({link})"
                    elif resultFormat=="mediawiki":
                        newValue=f"[{link} {value}]"
                    elif resultFormat=="latex":
                        newValue=f"\href{{{link}}}{{{value}}}"
                    if newValue is not None:
                        record[key]=newValue
                    
class QuerySyntaxHighlight:
    '''
    Syntax highlighting for queries with pygments
    '''
    def __init__(self,query,highlightFormat:str="html"):
        '''
        construct me for the given query and highlightFormat
        
        Args:
            query(Query): the query to do the syntax highlighting for
            highlightFormat(str): the highlight format to be used
        '''
        self.query=query
        self.highlightFormat=highlightFormat
        self.lexer=get_lexer_by_name(self.query.lang)
        if self.highlightFormat=="html":
            self.formatter=HtmlFormatter()
        elif self.highlightFormat=="latex":
            self.formatter=LatexFormatter()
        
    def highlight(self):
        '''
        Returns:
            str: the result of the syntax highlighting with pygments
        '''
        syntaxResult=highlight(self.query.query,self.lexer, self.formatter)    
        return syntaxResult

class QueryResultDocumentation():
    '''
    documentation of a query result
    '''
    
    def __init__(self,query,title:str,tablefmt:str,tryItMarkup:str,sourceCodeHeader:str,sourceCode:str,resultHeader:str,result:str):
        '''
        constructor
        
        Args:
            query(Query): the query to be documented
            title(str): the title markup
            tablefmt(str): the tableformat that has been used
            tryItMarkup: the "try it!" markup to show
            sourceCodeHeader(str): the header title to use for the sourceCode
            sourceCode(str): the sourceCode
            resultCodeHeader(str): the header title to use for the result
            result(str): the result header
            
        '''
        self.query=query
        self.title=title
        self.tablefmt=tablefmt
        self.tryItMarkup=f"\n{tryItMarkup}"
        self.sourceCodeHeader=sourceCodeHeader
        self.sourceCode=sourceCode
        self.resultHeader=resultHeader
        self.result=result
        
    @staticmethod
    def uniCode2Latex(text:str,withConvert:bool=False)->str:
        '''
        converts unicode text to latex and 
        fixes UTF-8 chars for latex in a certain range:
            ₀:$_0$ ... ₉:$_9$
            
        see https://github.com/phfaist/pylatexenc/issues/72
   
        Args:
            text(str): the string to fix
            withConvert(bool): if unicode to latex libary conversion should be used
        
        Return:
            str: latex presentation of UTF-8 char
        '''
        for code in range(8320,8330):
            text=text.replace(chr(code),f"$_{code-8320}$")
        if withConvert:
            latex=unicode_to_latex(text)
            # workaround {\textbackslash} being returned
            #latex=latex.replace("{\\textbackslash}",'\\')
            text=latex
        return text
        
    def __str__(self):
        '''
        simple string representation
        '''
        return self.asText()
        
    def asText(self):
        '''
        return my text representation
        
        Returns:
            str: description, sourceCodeHeader, sourceCode, tryIt link and result table
        '''
        text=f"{self.title}\n{self.query.description}\n{self.sourceCodeHeader}\n{self.sourceCode}{self.tryItMarkup}\n{self.resultHeader}\n{self.result}"
        fixedStr=self.uniCode2Latex(text) if self.tablefmt.lower()=="latex" else text
        return fixedStr

class Query(object):
    ''' a Query e.g. for SPAQRL '''
    
    def __init__(self,name:str,query:str,lang='sparql',endpoint:str=None,database:str="blazegraph",title:str=None,description:str=None,prefixes=None,tryItUrl:str=None,formats:list=None,debug=False):
        '''
        constructor 
        Args:
            name(string): the name/label of the query
            query(string): the native Query text e.g. in SPARQL
            lang(string): the language of the query e.g. SPARQL
            endpoint(string): the endpoint url to use 
            database(string): the type of database e.g. "blazegraph"
            title(string): the header/title of the query
            description(string): the description of the query
            prefixes(list): list of prefixes to be resolved
            tryItUrl(str): the url of a "tryit" webpage
            formats(list): key,value pairs of ValueFormatters to be applied
            debug(boolean): true if debug mode should be switched on
        '''
        self.name=name
        self.query=query
        self.lang=lang
        self.endpoint=endpoint
        self.database=database
        self.tryItUrl=tryItUrl
        
        self.title=title=name if title is None else title
        self.description="" if description is None else description
        self.prefixes=prefixes
        self.debug=debug
        self.formats=formats
        self.formatCallBacks=[]
        
    def __str__(self):
        queryStr="\n".join([f"{key}:{value}" for key, value in self.__dict__.items() if value is not None])
        return f"{queryStr}"
        
    def addFormatCallBack(self,callback):
        self.formatCallBacks.append(callback)
    
    def preFormatWithCallBacks(self,lod,tablefmt:str):
        '''
        run the configured call backs to pre-format the given list of dicts for the given tableformat
        
        Args:
            lod(list): the list of dicts to handle
            tablefmt(str): the table format (according to tabulate) to apply
        
        '''
        for record in lod:
            for key in record.keys():
                value=record[key]
                if value is not None:
                    for formatCallBack in self.formatCallBacks:
                        formatCallBack(record,key,value,tablefmt)
    
    def formatWithValueFormatters(self,lod,tablefmt:str):
        '''
        format the given list of Dicts with the ValueFormatters
        '''
        # is there anything to do?
        if self.formats is None:
            # no 
            return
        # get the value Formatters that might apply here
        valueFormatters=ValueFormatter.getFormats()
        formatsToApply={}
        for valueFormatSpec in self.formats:
            parts=valueFormatSpec.split(":")
            # e.g. president:wikidata
            keytoformat=parts[0]
            formatName=parts[1]
            if formatName in valueFormatters:
                formatsToApply[keytoformat]=valueFormatters[formatName]
        for record in lod:
            for keytoformat in formatsToApply:
                valueFormatter=formatsToApply[keytoformat]
                # format all key values
                if keytoformat=="*":
                    for key in record:
                        valueFormatter.applyFormat(record,key,tablefmt)
                # or just a selected one
                elif keytoformat in record:
                    valueFormatter.applyFormat(record,keytoformat,tablefmt)
            pass
        
    def getTryItUrl(self,baseurl:str,database:str="blazegraph"):
        '''
        return the "try it!" url for the given baseurl
        
        Args:
            baseurl(str): the baseurl to used
            
        Returns:
            str: the "try it!" url for the given query
        '''
        # https://stackoverflow.com/a/9345102/1497139
        quoted=urllib.parse.quote(str(self.query))
        if database=="blazegraph":
            delim="/#"
        else:
            delim="?query="
        url=f"{baseurl}{delim}{quoted}"
        return url
    
    def getLink(self,url,title,tablefmt):
        '''
        convert the given url and title to a link for the given tablefmt
        
        Args:
            url(str): the url to convert
            title(str): the title to show
            tablefmt(str): the table format to use
        '''
        # create a safe url
        if url is None:
            return ""
        markup=f"{title}:{url}"
        if tablefmt=="mediawiki":
            markup=f"[{url} {title}]"
        elif tablefmt=="github":
            markup=f"[{title}]({url})"
        elif tablefmt=="latex":
            markup=r"\href{%s}{%s}" % (url,title) 
        return markup
        
    def prefixToLink(self,lod:list,prefix:str,tablefmt:str):
        '''
        convert url prefixes to link according to the given table format
        TODO - refactor as preFormat callback
        
        Args:
            lod(list): the list of dicts to convert
            prefix(str): the prefix to strip 
            tablefmt(str): the tabulate tableformat to use
            
        '''
        for record in lod:
            for key in record.keys():
                value=record[key]
                if value is not None and isinstance(value,str) and value.startswith(prefix):
                    item=value.replace(prefix,"")
                    uqitem=urllib.parse.unquote(item)
                    if tablefmt=="latex":
                        link=uqitem
                    else:
                        link=self.getLink(value,uqitem,tablefmt)
                    record[key]=link
          
    def asYaml(self):
        yamlMarkup=yaml.dump(self)
        return yamlMarkup
    
    def asWikiSourceMarkup(self):
        '''
        convert me to Mediawiki markup for syntax highlighting using the "source" tag
        
        
        Returns:
            string: the Markup
        '''
        markup="<source lang='%s'>\n%s\n</source>\n" %(self.lang,self.query)
        return markup
        
    def asWikiMarkup(self,listOfDicts):
        '''
        convert the given listOfDicts result to MediaWiki markup
        
        Args:
            listOfDicts(list): the list of Dicts to convert to MediaWiki markup
        
        Returns:
            string: the markup
        '''
        if self.debug:
            print(listOfDicts)
        mwTable=MediaWikiTable()
        mwTable.fromListOfDicts(listOfDicts)
        markup=mwTable.asWikiMarkup()        
        return markup
    
    def documentQueryResult(self,qlod:list,limit=None,tablefmt:str="mediawiki",tryItUrl:str=None,withSourceCode=True,**kwArgs):
        '''
        document the given query results - note that a copy of the whole list is going to be created for being able to format
        
        Args:
            qlod: the list of dicts result
            limit(int): the maximum number of records to display in result tabulate
            tablefmt(str): the table format to use
            tryItUrl: the "try it!" url to show
            withSourceCode(bool): if True document the source code
            
        Return:
            str: the documentation tabular text for the given parameters
        '''
        sourceCode=self.query
        tryItMarkup=""
        sourceCodeHeader=""
        resultHeader=""
        title=self.title
        if limit is not None:
            lod=copy.deepcopy(qlod[:limit])
        else:
            lod=copy.deepcopy(qlod)    
        self.preFormatWithCallBacks(lod,tablefmt=tablefmt)
        self.formatWithValueFormatters(lod,tablefmt=tablefmt)
        result=tabulate(lod,headers="keys",tablefmt=tablefmt,**kwArgs)
        if tryItUrl is None and hasattr(self,'tryItUrl'):
            tryItUrl=self.tryItUrl
        if tablefmt=="github":
            title=f"## {self.title}"
            resultHeader="## result"
        elif tablefmt=="mediawiki":
            title=f"== {self.title} =="
            resultHeader="=== result ==="
        elif tablefmt=="latex":
            resultHeader=""
            result=r"""\begin{table}
            \caption{%s}
            \label{tab:%s}
            %s
            \end{table}
            """ % (self.title,self.name,result)
        else: 
            title=f"{self.title}"
            resultHeader="result:"
        if withSourceCode:
            tryItUrlEncoded=self.getTryItUrl(tryItUrl,self.database)
            tryItMarkup=self.getLink(tryItUrlEncoded, "try it!", tablefmt)
            if tablefmt=="github":
                sourceCodeHeader="### query"
                sourceCode=f"""```{self.lang}
{self.query}
```"""    
            elif tablefmt=="mediawiki":
                sourceCodeHeader="=== query ==="
                sourceCode=f"""<source lang='{self.lang}'>
{self.query}
</source>
"""
            elif tablefmt=="latex":
                sourceCodeHeader=r"see query listing \ref{listing:%s} and result table \ref{tab:%s}" % (self.name,self.name)
                sourceCode=r"""\begin{listing}[ht]
\caption{%s}
\label{listing:%s}
\begin{minted}{%s}
%s
\end{minted}
%s
\end{listing}
""" % (self.title,self.name,self.lang.lower(),self.query,tryItMarkup)
            else:
                sourceCodeHeader="query:"
                sourceCode=f"{self.query}"
        if self.lang!="sparql":
            tryItMarkup=""
        queryResultDocumentation=QueryResultDocumentation(query=self,title=title,tablefmt=tablefmt,tryItMarkup=tryItMarkup,sourceCodeHeader=sourceCodeHeader,sourceCode=sourceCode,resultHeader=resultHeader,result=result)
        return queryResultDocumentation

class QueryManager(object):
    '''
    manages pre packaged Queries
    '''

    def __init__(self,lang:str=None,debug=False,queriesPath=None):
        '''
        Constructor
        Args:
            lang(string): the language to use for the queries sql or sparql
            debug(boolean): True if debug information should be shown
        '''
        if lang is None:
            lang='sql'
        self.queriesByName={}
        self.lang=lang
        self.debug=debug
        queries=QueryManager.getQueries(queriesPath=queriesPath)
        for name,queryDict in queries.items():
            if self.lang in queryDict:
                queryText=queryDict.pop(self.lang)
                for qformat in ['sparql', 'sql']:   # drop not needed query variants
                    if qformat in queryDict:
                        queryDict.pop(qformat)
                query=Query(name=name,query=queryText,lang=self.lang,**queryDict,debug=self.debug)
                self.queriesByName[name]=query
    
    @staticmethod
    def getQueries(queriesPath=None):
        '''
        get the queries for thee given queries Path
        '''
        queriesPaths=YamlPath.getPaths("queries.yaml", queriesPath)
        queries={}
        for queriesPath in queriesPaths:
            with open(queriesPath, 'r') as stream:
                lqueries = yaml.safe_load(stream)
                for key in lqueries:
                    queries[key]=lqueries[key]
        return queries
            

class EndpointManager(object):
    """
    manages a set of SPARQL endpoints
    """

    @staticmethod
    def getEndpoints(endpointPath:str=None,lang:str=None):
        '''
        get the endpoints for the given endpointPath
        
        Args: 
            endpointPath(str): the path to the yaml file with the endpoint configurations
            lang(str): if lang is given filter by the given language
        '''
        endpointPaths=YamlPath.getPaths("endpoints.yaml",endpointPath)
        endpoints={}
        for lEndpointPath in endpointPaths:
            with open(lEndpointPath, 'r') as stream:
                endpointRecords = yaml.safe_load(stream)
                for name, record in endpointRecords.items():
                    select=True
                    if lang is not None:
                        select=record["lang"]==lang
                    if select:
                        endpoint=Endpoint()
                        endpoint.fromDict({"name": name, **record})
                        endpoints[name]=endpoint
        return endpoints

    @staticmethod
    def getEndpointNames(endpointPath=None) -> list:
        """
        Returns a list of all available endpoint names
        """
        endpoints = EndpointManager.getEndpoints(endpointPath)
        return list(endpoints.keys())


class Endpoint(JSONAble):
    """
    a query endpoint
    """

    @staticmethod
    def getSamples():
        samples=[
            {
                "name": "wikidata",
                "lang": "sparql",
                "endpoint": "https://query.wikidata.org/sparql",
                "website": "https://query.wikidata.org/",
                "database": "blazegraph",
                "method": "POST",
                "prefixes": "PREFIX bd: <http://www.bigdata.com/rdf#>\nPREFIX cc: <http://creativecommons.org/ns#>"
            }
        ]
        return samples
    
    @classmethod
    def getDefault(cls):
        endpointConf=Endpoint()
        endpointConf.fromDict(Endpoint.getSamples()[0])
        return endpointConf
    
    def __init__(self):
        '''
        constructor for setting defaults
        '''
        self.method="POST"
        self.lang="SPARQL"
        
    
    def __str__(self):
        '''
        Returns:    
            str: a string representation of this Endpoint
        '''
        text=f"{self.name}:{self.website}:{self.endpoint}({self.method})"
        return text


'''
Created on 2022-04-14

@author: wf
'''
import datetime
from lodstorage.sparql import SPARQL
from lodstorage.query import Query,QueryManager,YamlPath,Endpoint
from lodstorage.version import Version
from pathlib import Path
import os
import textwrap
import re


class Variable:
    '''
    Variable e.g. name handling
    '''
    @classmethod
    def validVarName(cls,varStr:str)->str:
        '''
        convert the given potential variable name string to a valid  
        variable name
        
        see https://stackoverflow.com/a/3305731/1497139
        
        Args:
            varStr(str): the string to convert
            
        Returns:
            str: a valid variable name
        '''
        return re.sub('\W|^(?=\d)','_', varStr)
    
class WikidataProperty():
    '''
    a WikidataProperty
    '''
    
    def __init__(self,pid:str):
        '''
        construct me with the given property id
        
        Args:
            pid(str): the property Id
        '''
        self.pid=pid
        self.reverse=False
    
    def getPredicate(self):
        '''
        get me as a Predicate
        '''
        reverseToken="^" if self.reverse else ""
        plabel=f"{reverseToken}wdt:{self.pid}"
        return plabel
    
    def __str__(self):
        text=self.pid
        if hasattr(self, "plabel"):
            text=f"{self.plabel} ({self.pid})"
        return text
      
    @classmethod
    def getPropertiesByLabels(cls,sparql,propertyLabels:list,lang:str="en"):
        '''
        get a list of Wikidata properties by the given label list
        
        Args:
            sparql(SPARQL): the SPARQL endpoint to use
            propertyLabels(list): a list of labels of the properties 
            lang(str): the language of the label
        '''
        # the result dict
        wdProperties={}
        if len(propertyLabels)>0:
            valuesClause=""
            for propertyLabel in propertyLabels:
                valuesClause+=f'   "{propertyLabel}"@{lang}\n'
            query=f"""# get the properties for the given labels
{WikidataItem.getPrefixes(["rdf","rdfs","wikibase"])}
SELECT ?property ?propertyLabel ?wbType WHERE {{
  VALUES ?propertyLabel {{
{valuesClause}
  }}
  ?property rdf:type wikibase:Property;rdfs:label ?propertyLabel.
  ?property wikibase:propertyType  ?wbType.
  FILTER(LANG(?propertyLabel) = "{lang}")
}}"""
            cls.addPropertiesForQuery(wdProperties,sparql,query)
        return wdProperties
    
    @classmethod
    def getPropertiesByIds(cls,sparql,propertyIds:list,lang:str="en"):
        '''
        get a list of Wikidata properties by the given id list
        
        Args:
            sparql(SPARQL): the SPARQL endpoint to use
            propertyIds(list): a list of ids of the properties 
            lang(str): the language of the label
        '''
        # the result dict
        wdProperties={}
        if len(propertyIds)>0:
            valuesClause=""
            for propertyId in propertyIds:
                valuesClause+=f'   wd:{propertyId}\n'
            query=f"""
# get the property for the given property Ids
{WikidataItem.getPrefixes(["rdf","rdfs","wd","wikibase"])}
SELECT ?property ?propertyLabel ?wbType WHERE {{
  VALUES ?property {{
{valuesClause}
  }}
  ?property rdf:type wikibase:Property;rdfs:label ?propertyLabel.
  ?property wikibase:propertyType  ?wbType.
  FILTER(LANG(?propertyLabel) = "{lang}")
}}""" 
            cls.addPropertiesForQuery(wdProperties,sparql,query)
        return wdProperties
        
    @classmethod    
    def addPropertiesForQuery(cls,wdProperties:list,sparql,query):  
        '''
          add properties from the given query's result to the given
          wdProperties list using the given sparql endpoint
        Args:
          wdProperties(list): the list of wikidata properties
          sparql(SPARQL): the SPARQL endpoint to use
          query(str): the SPARQL query to perform
        '''
        qLod=sparql.queryAsListOfDicts(query)
        for record in qLod:
            url=record["property"]
            pid=re.sub(r"http://www.wikidata.org/entity/(.*)",r"\1",url)
            prop=WikidataProperty(pid)
            prop.plabel=record["propertyLabel"]
            prop.wbtype=record["wbType"]
            prop.url=url
            wdProperties[prop.plabel]=prop
            prop.varname=Variable.validVarName(prop.plabel)
            pass
        return wdProperties
    
class WikidataItem:
    '''
    a wikidata Item
    '''
    def __init__(self,qid:str,lang:str="en",sparql:SPARQL=None):
        '''
        construct me with the given item id, language and optional SPARQL access
        
        Args:
            qid(str): the item Id
            lang(str): the language to use
            sparql(SPARQL): the sparql access to use
        '''
        self.qid=qid
        # numeric qid
        self.qnumber=int(qid[1:])
        self.url=f"https://www.wikidata.org/wiki/{self.qid}"
        self.lang=lang
        self.sparql=sparql
        if sparql is not None:
            self.qlabel,self.description=WikidataItem.getLabelAndDescription(sparql, self.qid, self.lang)
            self.varname=Variable.validVarName(self.qlabel)
    
    def __str__(self):
        return self.asText(long=False)
    
    def asText(self,long:bool=True,wrapAt:int=0):
        '''
        returns my content as a text representation
        
        Args:
            long(bool): True if a long format including url is wished
            wrapAt(int): wrap long lines at the given width (if >0)
            
        Returns:
            str: a text representation of my content
        '''
        text=self.qid
        if hasattr(self, "qlabel"):
            text=f"{self.qlabel} ({self.qid})"  
        if hasattr(self,"description"):
            desc=self.description
            if wrapAt>0:
                desc=textwrap.fill(desc,width=wrapAt)
            text+=f"☞{desc}"
        if long:
            text+=f"→ {self.url}"
        return text
    
    @classmethod
    def getPrefixes(cls,prefixes=["rdf","rdfs","schema","wd","wdt","wikibase","xsd"]):
        prefixMap={
          "bd":"<http://www.bigdata.com/rdf#>",
          "cc":"<http://creativecommons.org/ns#>",
          "dct":"<http://purl.org/dc/terms/>",
          "geo":"<http://www.opengis.net/ont/geosparql#>",
          "ontolex":"<http://www.w3.org/ns/lemon/ontolex#>",
          "owl":"<http://www.w3.org/2002/07/owl#>",
          "p":"<http://www.wikidata.org/prop/>",
          "pq":"<http://www.wikidata.org/prop/qualifier/>",
          "pqn":"<http://www.wikidata.org/prop/qualifier/value-normalized/>",
          "pqv":"<http://www.wikidata.org/prop/qualifier/value/>",
          "pr":"<http://www.wikidata.org/prop/reference/>",
          "prn":"<http://www.wikidata.org/prop/reference/value-normalized/>",
          "prov":"<http://www.w3.org/ns/prov#>",
          "prv":"<http://www.wikidata.org/prop/reference/value/>",
          "ps":"<http://www.wikidata.org/prop/statement/>",
          "psn":"<http://www.wikidata.org/prop/statement/value-normalized/>",
          "psv":"<http://www.wikidata.org/prop/statement/value/>",
          "rdf":"<http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
          "rdfs":"<http://www.w3.org/2000/01/rdf-schema#>",
          "schema":"<http://schema.org/>",
          "skos":"<http://www.w3.org/2004/02/skos/core#>",
          "wd":"<http://www.wikidata.org/entity/>",
          "wdata":"<http://www.wikidata.org/wiki/Special:EntityData/>",
          "wdno":"<http://www.wikidata.org/prop/novalue/>",
          "wdref":"<http://www.wikidata.org/reference/>",
          "wds":"<http://www.wikidata.org/entity/statement/>",
          "wdt":"<http://www.wikidata.org/prop/direct/>",
          "wdtn":"<http://www.wikidata.org/prop/direct-normalized/>",
          "wdv":"<http://www.wikidata.org/value/>",
          "wikibase":"<http://wikiba.se/ontology#>",
          "xsd":"<http://www.w3.org/2001/XMLSchema#>"
        }
        # see also https://www.wikidata.org/wiki/EntitySchema:E49
        sparql=""
        for prefix in prefixes:
            if prefix in prefixMap:
                sparql+=f"PREFIX {prefix}: {prefixMap[prefix]}\n"
        return sparql

    @classmethod
    def getLabelAndDescription(cls,sparql:SPARQL, itemId:str,lang:str="en"):
        '''
        get  the label for the given item and language
        
        Args:
            itemId(str): the wikidata Q/P id
            lang(str): the language of the label 
            
        Returns:
            (str,str): the label and description as a tuple
        '''
        query=f"""# get the label for the given item
{cls.getPrefixes(["rdfs","wd","schema"])}        
SELECT ?itemLabel ?itemDescription
WHERE
{{
  VALUES ?item {{
    wd:{itemId}
  }}
  ?item rdfs:label ?itemLabel.
  FILTER (LANG(?itemLabel) = "{lang}").
  ?item schema:description ?itemDescription.
  FILTER(LANG(?itemDescription) = "{lang}")
}}""" 
        return sparql.getValues(query, ["itemLabel","itemDescription"])
        
    @classmethod
    def getItemsByLabel(cls,sparql:SPARQL,itemLabel:str,lang:str="en")->list:
        '''
        get a Wikidata items by the given label
        
        Args:
            sparql(SPARQL): the SPARQL endpoint to use
            itemLabel(str): the label of the items
            lang(str): the language of the label
            
        Returns:
            a list of potential items
        '''
        valuesClause=f'   "{itemLabel}"@{lang}\n'
        query=f"""# get the items that have the given label in the given language
# e.g. we'll find human=Q5 as the oldest type for the label "human" first
# and then the newer ones such as "race in Warcraft"
{cls.getPrefixes(["rdfs","schema","xsd"])}
SELECT 
  #?itemId 
  ?item 
  ?itemLabel 
  ?itemDescription
WHERE {{ 
  VALUES ?itemLabel {{
    {valuesClause}
  }}
  #BIND (xsd:integer(SUBSTR(STR(?item),33)) AS ?itemId)
  ?item rdfs:label ?itemLabel. 
  ?item schema:description ?itemDescription.
  FILTER(LANG(?itemDescription)="{lang}")
}} 
#ORDER BY ?itemId""" 
        qLod=sparql.queryAsListOfDicts(query)
        items=[]
        for record in qLod:
            url=record["item"] 
            qid=re.sub(r"http://www.wikidata.org/entity/(.*)",r"\1",url)
            item=WikidataItem(qid)
            item.url=url
            item.qlabel=record["itemLabel"]
            item.varname=Variable.validVarName(item.qlabel)
            item.description=record["itemDescription"]
            items.append(item)
        sortedItems=sorted(items,key=lambda item: item.qnumber)
        return sortedItems
        
class TrulyTabular(object):
    '''
    truly tabular SPARQL/RDF analysis
    
    checks "how tabular" a query based on a list of properties of an itemclass is
    '''

    def __init__(self, itemQid, propertyLabels:list=[],propertyIds:list=[],where:str=None,endpointConf=None,lang="en",debug=False):
        '''
        Constructor
        
        Args:
            itemQid(str): wikidata id of the type to analyze 
            propertyLabels(list): a list of labels of properties to be considered
            propertyIds(list): a list of ids of properties to be considered
            where(str): extra where clause for instance selection (if any)
            endpoint(str): the url of the SPARQL endpoint to be used
        '''
        self.itemQid=itemQid
        self.debug=debug
        if endpointConf is None:
            endpointConf=Endpoint.getDefault()
        self.endpointConf=endpointConf
        self.sparql=SPARQL(endpointConf.endpoint,method=self.endpointConf.method)
        self.sparql.debug=self.debug
        self.where=f"\n  {where}" if where is not None else ""
        self.lang=lang
        self.item=WikidataItem(itemQid,sparql=self.sparql,lang=lang)
        self.queryManager=TrulyTabular.getQueryManager(debug=self.debug)
        self.properties=WikidataProperty.getPropertiesByIds(self.sparql,propertyIds,lang)
        self.properties.update(WikidataProperty.getPropertiesByLabels(self.sparql, propertyLabels, lang))
        self.isodate=datetime.datetime.now().isoformat()
        self.error=None
        
    def __str__(self):
        '''
        Returns:
            str: my text representation
        '''
        return self.asText(long=False)
    
    def count(self):
        '''
        get my count
        '''
        itemText=self.getItemText()
        query=f"""# Count all items with the given type
# {itemText}
{WikidataItem.getPrefixes()}
SELECT (COUNT (DISTINCT ?item) AS ?count)
WHERE
{{
  # instance of {self.item.qlabel}
  ?item wdt:P31 wd:{self.item.qid}.{self.where}
}}"""
        try:
            count=self.sparql.getValue(query, "count")
            # workaround https://github.com/ad-freiburg/qlever/issues/717
            count=int(count)
        except Exception as ex:
            self.error=ex
            count=None
            
        return count,query
    
    def asText(self,long:bool=True):
        '''
        returns my content as a text representation
        
        Args:
            long(bool): True if a long format including url is wished
            
        Returns:
            str: a text representation of my content
        '''
        text=self.item.asText(long)
        return text
    
    def getItemText(self):
        # leads to 405 Method not allowed in SPARQLWrapper under certain circumstances
        # itemText=self.asText(long=True)
        itemText=f"{self.itemQid}:{self.item.qlabel}"
        return itemText
    
    @classmethod
    def getQueryManager(cls,lang='sparql',name="trulytabular",debug=False):
        '''
        get the query manager for the given language and fileName
        
        Args:
            lang(str): the language of the queries to extract
            name(str): the name of the manager containing the query specifications
            debug(bool): if True set debugging on
        '''
        qYamlFileName=f"{name}.yaml"
        for qYamlFile in YamlPath.getPaths(qYamlFileName):
            if os.path.isfile(qYamlFile):
                qm=QueryManager(lang=lang,debug=debug,queriesPath=qYamlFile)
                return qm
        return None
    
    def generateSparqlQuery(self,genMap:dict,listSeparator:str="⇹",naive:bool=True,lang:str='en')->str:
        '''
        generate a SPARQL Query
        
        Args:
            genMap(dict): a dictionary of generation items aggregates/ignores/labels
            listSeparator(str): the symbole to use as a list separator for GROUP_CONCAT
            naive(bool): if True - generate a naive straight forward SPARQL query
                if False generate a proper truly tabular aggregate query
            lang(str): the language to generate for
            
        Returns:
            str: the generated SPARQL Query
        '''
        # The Wikidata item to generate the query for
        item=self.item
        # the name of this script
        script=Path(__file__).name
        # the mode of generation
        naiveText="naive" if naive else "aggregate"
        # start with th preamble and PREFIX section
        # select the item and it's label
        sparqlQuery=f"""# truly tabular {naiveText} query for 
# {item.qid}:{item.qlabel}
# generated by {script} version {Version.version} on {self.isodate}
{WikidataItem.getPrefixes()}
SELECT ?{item.varname} ?{item.varname}Label"""
        # loop over all properties
        for wdProp in self.properties.values():
            if naive:
                sparqlQuery+=f"\n  ?{wdProp.varname}"
            else:
                if wdProp.pid in genMap:
                    genList=genMap[wdProp.pid]
                    for aggregate in genList:
                        if not aggregate in ["ignore","label"]:
                            distinct=""
                            if aggregate=="list": 
                                aggregateFunc="GROUP_CONCAT"
                                aggregateParam=f';SEPARATOR="{listSeparator}"'
                                distinct="DISTINCT "
                            else:
                                if aggregate=="count":
                                    distinct="DISTINCT "
                                aggregateFunc=aggregate.upper()
                                aggregateParam=""
                            sparqlQuery+=f"\n  ({aggregateFunc} ({distinct}?{wdProp.varname}{aggregateParam}) AS ?{wdProp.varname}_{aggregate})"
                        elif aggregate=="label":
                            sparqlQuery+=f"\n  ?{wdProp.varname}Label"
                        elif aggregate=="ignore":
                            sparqlQuery+=f"\n  ?{wdProp.varname}"    
        sparqlQuery+=f"""
WHERE {{
  # instanceof {item.qid}:{item.qlabel}
  ?{item.varname} wdt:P31 wd:{item.qid}.
  # label
  ?{item.varname} rdfs:label ?{item.varname}Label.  
  FILTER (LANG(?{item.varname}Label) = "{lang}").
"""
        for wdProp in self.properties.values():
            sparqlQuery+=f"""  # {wdProp}
  OPTIONAL {{ 
    ?{item.varname} wdt:{wdProp.pid} ?{wdProp.varname}. """
            if wdProp.pid in genMap:
                genList=genMap[wdProp.pid]
                if "label" in genList:
                    sparqlQuery+=f"""\n    ?{wdProp.varname} rdfs:label ?{wdProp.varname}Label."""
                    sparqlQuery+=f"""\n    FILTER (LANG(?{wdProp.varname}Label) = "{lang}")."""
            sparqlQuery+="\n  }\n"
        # close where Clause
        sparqlQuery+="""}\n"""
        # optionally add Aggregate
        if not naive:
            sparqlQuery+=f"""GROUP BY
  ?{item.varname} 
  ?{item.varname}Label
"""
            for wdProp in self.properties.values():
                if wdProp.pid in genMap:
                    genList=genMap[wdProp.pid]
                    if "label" in genList:
                        sparqlQuery+=f"\n  ?{wdProp.varname}Label"
            havingCount=0
            havingDelim="   "
            for wdProp in self.properties.values():
                if wdProp.pid in genMap:
                    genList=genMap[wdProp.pid]
                    if "ignore" in genList:
                        havingCount+=1
                        if havingCount==1:
                            sparqlQuery+=f"\nHAVING ("
                            
                        sparqlQuery+=f"\n  {havingDelim}COUNT(?{wdProp.varname})<=1"
                        havingDelim="&& "
            if havingCount>0:
                sparqlQuery+=f"\n)"
        return sparqlQuery
    
    def mostFrequentPropertiesQuery(self,whereClause:str=None,minCount:int=0):
        '''
        get the most frequently used properties
        
        Args:
            whereClause(str): an extra WhereClause to use
        '''
        if whereClause is None:
            whereClause=f"?item wdt:P31 wd:{self.itemQid}";
            if self.endpointConf.database!="qlever":
                whereClause+=";?p ?id"
        whereClause+="."  
        minCountFilter=""
        if minCount>0:
            minCountFilter=f"\n  FILTER(?count >{minCount})."
        itemText=self.getItemText()
        sparqlQuery=f"""# get the most frequently used properties for
# {itemText}
{WikidataItem.getPrefixes()}
SELECT ?prop ?propLabel ?wbType ?count WHERE {{
  {{"""
        if self.endpointConf.database=="qlever":
            sparqlQuery+=f"""
    SELECT ?p (COUNT(DISTINCT ?item) AS ?count) WHERE {{"""
        else:
            sparqlQuery+=f"""
    SELECT ?prop (COUNT(DISTINCT ?item) AS ?count) WHERE {{"""
        if self.endpointConf.database=="blazegraph":
            sparqlQuery+=f"""
      hint:Query hint:optimizer "None"."""
        sparqlQuery+=f"""
      {whereClause}"""
        if self.endpointConf.database=="qlever":
            sparqlQuery+=f"""  
      ?item ql:has-predicate ?p 
    }} GROUP BY ?p
  }}
  ?prop wikibase:directClaim ?p."""
        else:
            sparqlQuery+=f"""
      ?prop wikibase:directClaim ?p.
    }}
    GROUP BY ?prop ?propLabel
  }}"""
        sparqlQuery+=f"""
  ?prop rdfs:label ?propLabel.
  ?prop wikibase:propertyType ?wbType.
  FILTER(LANG(?propLabel) = "{self.lang}").{minCountFilter}  
}}
ORDER BY DESC (?count)
"""
        title=f"most frequently used properties for {self.item.asText(long=True)}"
        query=Query(name=f"mostFrequentProperties for {itemText}",query=sparqlQuery,title=title)
        return query
    
    def noneTabularQuery(self,wdProperty:WikidataProperty,asFrequency:bool=True):
        '''
        get the none tabular entries for the given property
        
        Args:
            wdProperty(WikidataProperty): the property to analyze
            asFrequency(bool): if true do a frequency analysis
        '''
        propertyLabel=wdProperty.plabel
        propertyId=wdProperty.pid
        # work around https://github.com/RDFLib/sparqlwrapper/issues/211
        if "described at" in propertyLabel:
            propertyLabel=propertyLabel.replace("described at","describ'd at")
        sparql=f"""SELECT ?item ?itemLabel (COUNT (?value) AS ?count)
WHERE
{{
  # instance of {self.item.qlabel}
  ?item wdt:P31 wd:{self.itemQid}.{self.where}
  ?item rdfs:label ?itemLabel.
  FILTER (LANG(?itemLabel) = "{self.lang}").
  # {propertyLabel}
  ?item {wdProperty.getPredicate()} ?value.
}} GROUP BY ?item ?itemLabel
"""
        if asFrequency:
            freqDesc="frequencies"
            sparql=f"""SELECT ?count (COUNT(?count) AS ?frequency) WHERE {{{{
{sparql}
}}}}
GROUP BY ?count
ORDER BY DESC (?frequency)"""
        else:
            freqDesc="records"
            sparql=f"""{sparql}
HAVING (COUNT (?value) > 1)
ORDER BY DESC(?count)"""
        itemText=self.getItemText()
        sparql=f"""# Count all {itemText} items
# with the given {propertyLabel}({propertyId}) https://www.wikidata.org/wiki/Property:{propertyId} 
{WikidataItem.getPrefixes()}
"""+sparql
        title=f"non tabular entries for {self.item.qlabel}/{propertyLabel}:{freqDesc}"
        name=f"NonTabular {self.item.qlabel}/{propertyLabel}:{freqDesc}"
        query=Query(query=sparql,name=name,title=title)
        return query

    def noneTabular(self,wdProperty:WikidataProperty):
        '''
        get the none tabular result for the given Wikidata property
        
        Args:
            wdProperty(WikidataProperty): the Wikidata property
        '''
        query=self.noneTabularQuery(wdProperty)
        if self.debug:
            print(query.query)
        qlod=self.sparql.queryAsListOfDicts(query.query)
        return qlod
    
    def addStatsColWithPercent(self,m,col,value,total): 
        '''
        add a statistics Column
        '''
        m[col]=value
        m[f"{col}%"]=float(f"{value/total*100:.1f}") 
        
    def genWdPropertyStatistic(self,wdProperty:WikidataProperty,itemCount:int,withQuery=True)->dict:
        '''
        generate a property Statistics Row for the given wikidata Property
        
        Args:
            wdProperty(WikidataProperty): the property to get the statistics for
            itemCount(int): the total number of items to check
            withQuery(bool): if true include the sparql query
            
        Returns:
            dict: a statistics row
        '''
        ntlod=self.noneTabular(wdProperty)
        statsRow={"property":wdProperty.plabel}
        total=0
        nttotal=0
        maxCount=0
        for record in ntlod:
            f=int(record["frequency"])
            count=int(record["count"])
            #statsRow[f"f{count}"]=f
            if count>1:
                nttotal+=f
            else:
                statsRow["1"]=f
            if count>maxCount:
                maxCount=count     
            total+=f
        statsRow["maxf"]=maxCount
        if withQuery:
            statsRow["queryf"]=self.noneTabularQuery(wdProperty).query
            statsRow["queryex"]=self.noneTabularQuery(wdProperty,asFrequency=False).query
        self.addStatsColWithPercent(statsRow,"total",total,itemCount)
        self.addStatsColWithPercent(statsRow,"non tabular",nttotal,total)
        return statsRow
        
    def genPropertyStatistics(self):
        '''
        generate the property Statistics
        
        Returns:
            generator: a generator of statistic dict rows
        '''
        itemCount,_itemCountQuery=self.count()
        for wdProperty in self.properties.values():
            statsRow=self.genWdPropertyStatistic(wdProperty, itemCount)
            yield statsRow
    
    def getPropertyStatistics(self):
        '''
        get the property Statistics
        '''
        itemCount,_itemCountQuery=self.count()
        lod=[{
            "property": "∑",
            "total": itemCount,
            "total%": 100.0
        }]
        for wdProperty in self.properties.values():
            statsRow=self.genWdPropertyStatistic(wdProperty, itemCount)
            lod.append(statsRow)
        return lod
    

        


'''
Created on 2021-01-31

@author: wf
'''

class LOD(object):
    '''
    list of Dict aka Table
    '''

    def __init__(self,name):
        '''
        Constructor
        '''
        self.name=name
        pass
    
    @staticmethod
    def getFields(listOfDicts,sampleCount:int=None):
        if sampleCount is None:
            if listOfDicts is None:
                return None
            sampleCount=len(listOfDicts)
        fields=[]
        from lodstorage.jsonable import JSONAble
        for row in listOfDicts:
            if isinstance(row, JSONAble):
                row=vars(row)
            for key in row.keys():
                if not key in fields:
                    fields.append(key)
        return fields
        
    @staticmethod    
    def setNone4List(listOfDicts,fields):
        '''
        set the given fields to None for the records in the given listOfDicts
        if they are not set
        Args:
            listOfDicts(list): the list of records to work on
            fields(list): the list of fields to set to None 
        '''
        for record in listOfDicts:
            LOD.setNone(record, fields)
    
    @staticmethod
    def setNone(record,fields):
        '''
        make sure the given fields in the given record are set to none
        Args:
            record(dict): the record to work on
            fields(list): the list of fields to set to None 
        '''
        for field in fields:
            if not field in record:
                record[field]=None
                
    '''
    https://stackoverflow.com/questions/33542997/python-intersection-of-2-lists-of-dictionaries/33543164
    '''
    @staticmethod  
    def sortKey(d,key=None):
        ''' get the sort key for the given dict d with the given key
        '''
        if key is None:
            # https://stackoverflow.com/a/60765557/1497139
            return hash(tuple(d.items()))
        else:
            return d[key] 
        
    @staticmethod            
    def intersect(listOfDict1,listOfDict2,key=None):
        '''
        get the  intersection of the two lists of Dicts by the given key 
        '''
        i1=iter(sorted(listOfDict1, key=lambda k: LOD.sortKey(k, key)))
        i2=iter(sorted(listOfDict2, key=lambda k: LOD.sortKey(k, key)))
        c1=next(i1)
        c2=next(i2)
        lr=[]
        while True:
            try:
                val1=LOD.sortKey(c1,key)
                val2=LOD.sortKey(c2,key)
                if val1<val2:
                    c1=next(i1)
                elif val1>val2:
                    c2=next(i2)
                else:
                    lr.append(c1)
                    c1=next(i1)
                    c2=next(i2)
            except StopIteration:
                break     
        return lr    
    
    @staticmethod
    def addLookup(lookup,duplicates,record,value,withDuplicates:bool):
        '''
        add a single lookup result
        
        Args:
            lookup(dict): the lookup map
            duplicates(list): the list of duplicates
            record(dict): the current record
            value(object): the current value to lookup
            withDuplicates(bool): if True duplicates should be allowed and lists returned if False a separate duplicates
            list is created 
        '''
        if value in lookup:
            if withDuplicates:
                lookupResult=lookup[value]
                lookupResult.append(record)
            else:
                duplicates.append(record)
                return
        else:
            if withDuplicates:
                lookupResult=[record]
            else:
                lookupResult=record
        lookup[value]=lookupResult  
    
    @staticmethod
    def getLookup(lod:list,attrName:str,withDuplicates:bool=False):
        '''
        create a lookup dictionary by the given attribute name for the given list of dicts
        
        Args:
            lod(list): the list of dicts to get the lookup dictionary for
            attrName(str): the attribute to lookup
            withDuplicates(bool): whether to retain single values or lists
        
        Return:
            a dictionary for lookup
        '''
        lookup={}
        duplicates=[]
        for record in lod:
            value=None
            if isinstance(record,dict):
                if attrName in record:
                    value=record[attrName]
            else:
                if hasattr(record, attrName):
                    value=getattr(record,attrName)
            if value is not None:
                if isinstance(value,list):
                    for listValue in value:
                        LOD.addLookup(lookup,duplicates,record,listValue,withDuplicates)    
                else:
                    LOD.addLookup(lookup,duplicates,record,value,withDuplicates)    
        if withDuplicates:
            return lookup  
        else:
            return lookup,duplicates
    
    @classmethod
    def handleListTypes(cls,lod,doFilter=False,separator=","):
        '''
        handle list types in the given list of dicts
        
        Args:
            cls: this class
            lod(list): a list of dicts
            doFilter(bool): True if records containing lists value items should be filtered
            separator(str): the separator to use when converting lists
        '''
        # see https://stackoverflow.com/a/1207485/1497139
        for i in range(len(lod) - 1, -1, -1):
            record=lod[i]
            if isinstance(record,dict):
                for key in record:
                    value=record[key]
                    if isinstance(value,list):
                        if doFilter:
                            del lod[i]
                            continue
                        else:
                            newValue=separator.join(filter(None, value))
                            record[key]=newValue
        
    @staticmethod
    def filterFields(lod:list, fields:list, reverse:bool=False):
        '''
        filter the given LoD with the given list of fields by either limiting the LoD to the fields or removing the
        fields contained in the list depending on the state of the reverse parameter

        Args:
            lod(list): list of dicts from which the fields should be excluded
            fields(list): list of fields that should be excluded from the lod
            reverse(bool): If True limit dict to the list of given fields. Otherwise exclude the fields from the dict.

        Returns:
            LoD
        '''
        res=[]
        for record in lod:
            if reverse:
                recordReduced={d: record[d] for d in record if d in fields}
            else:
                recordReduced = {d: record[d] for d in record if d not in fields}
            res.append(recordReduced)
        return res


# ! important
# see https://stackoverflow.com/a/27868004/1497139
from setuptools import setup
from collections import OrderedDict
from lodstorage.version import Version
import pathlib

# where is this setup file?
here = pathlib.Path(__file__).parent.resolve()
# get thre requirements
requirements = (here / 'requirements.txt').read_text(encoding='utf-8').split("\n")
# and the README content
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name=Version.name,
    version=Version.version,
    description=Version.description,
    packages=['lodstorage','sampledata'],
    package_data={
          'sampledata': ['*.yaml'],
    },
    author='Wolfgang Fahl',
    author_email='wf@bitplan.com',
    maintainer='Wolfgang Fahl',
    url='https://github.com/WolfgangFahl/pyLodStorage',
    project_urls=OrderedDict(
        (
            ("Documentation", "http://wiki.bitplan.com/index.php/PyLoDStorage"),
            ("Code", "https://github.com/WolfgangFahl/pyLoDStorage/blob/master/lodstorage/sql.py"),
            ("Issue tracker", "https://github.com/WolfgangFahl/pyLodStorage/issues"),
        )
    ),
    license='Apache License',
   
    install_requires=requirements,
    classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10'
    ],
    entry_points={
      'console_scripts': [
        'sparqlquery = lodstorage.querymain:mainSPARQL',
        'sqlquery=lodstorage.querymain:mainSQL',
        'lodquery=lodstorage.querymain:main'
      ],
    },
    long_description=long_description,
    long_description_content_type='text/markdown'
)



'''
Created on 2020-08-14

@author: wf
'''
from SPARQLWrapper import SPARQLWrapper2
from SPARQLWrapper.Wrapper import POSTDIRECTLY, POST
from lodstorage.lod import LOD
import datetime
import time
from sys import stderr

class SPARQL(object):
    '''
    wrapper for SPARQL e.g. Apache Jena, Virtuoso, Blazegraph
    
    :ivar url: full endpoint url (including mode)
    :ivar mode: 'query' or 'update'
    :ivar debug: True if debugging is active
    :ivar typedLiterals: True if INSERT should be done with typedLiterals
    :ivar profile(boolean): True if profiling / timing information should be displayed
    :ivar sparql: the SPARQLWrapper2 instance to be used
    :ivar method(str): the HTTP method to be used 'POST' or 'GET'
    '''
    def __init__(self,url,mode='query', debug=False, isFuseki=False,typedLiterals=False,  profile=False, agent='PyLodStorage',method='POST'):
        '''
        Constructor a SPARQL wrapper
        
        Args:
            url(string): the base URL of the endpoint - the mode query/update is going to be appended
            mode(string): 'query' or 'update'
            debug(bool): True if debugging is to be activated
            typedLiterals(bool): True if INSERT should be done with typedLiterals
            profile(boolean): True if profiling / timing information should be displayed
            agent(string): the User agent to use
            method(string): the HTTP method to be used 'POST' or 'GET'
        '''
        if isFuseki:
            self.url=f"{url}/{mode}"
        else:
            self.url=url
        self.mode=mode
        self.debug=debug
        self.typedLiterals=typedLiterals
        self.profile=profile
        self.sparql=SPARQLWrapper2(url)
        self.method=method
        self.sparql.agent=agent
        
    def rawQuery(self,queryString,method='POST'):
        '''
        query with the given query string
        
        Args:
            queryString(string): the SPARQL query to be performed
            method(string): POST or GET - POST is mandatory for update queries
        Returns:
            list: the raw query result as bindings
        '''
        self.sparql.setQuery(queryString)
        self.sparql.method=method
        queryResult = self.sparql.query()
        return queryResult
    
    def getValue(self,sparqlQuery:str,attr:str):
        '''
        get the value for the given SPARQL query using the given attr
        
        Args:
            sparql(SPARQL): the SPARQL endpoint to ge the value for
            sparqlQuery(str): the SPARQL query to run
            attr(str): the attribute to get
        '''
        if self.debug:
            print(sparqlQuery)
        qLod=self.queryAsListOfDicts(sparqlQuery)
        return self.getFirst(qLod, attr)
    
    def getValues(self,sparqlQuery:str,attrList:list):
        '''
        get Values for the given sparlQuery and attribute list
        '''
        if self.debug:
            print(sparqlQuery)
        qLod=self.queryAsListOfDicts(sparqlQuery)
        if not (len(qLod)==1):
            msg=f"getValues for {attrList} failed for {qLod}"
            raise Exception(msg)
        record=qLod[0]
        values=()
        for attr in attrList:
            if not attr in record:
                msg=f"getValues failed for attribute {attr} which is missing in result record {record}"
                raise Exception(msg)
            recordTuple=(record[attr],)
            values+=recordTuple
        return values
    
    def getFirst(self,qLod:list,attr:str):
        '''
        get the column attr of the first row of the given qLod list
        
        Args:
            qLod(list): the list of dicts (returned by a query)
            attr(str): the attribute to retrieve
            
        Returns:
            object: the value
        '''
        if len(qLod)==1 and attr in qLod[0]:
            value=qLod[0][attr]
            return value
        raise Exception(f"getFirst for attribute {attr} failed for {qLod}")
    
    def getResults(self,jsonResult):
        '''
        get the result from the given jsonResult
        
        Args:
            jsonResult: the JSON encoded result
            
        Returns:
            list: the list of bindings
        '''    
        return jsonResult.bindings
    
    def insert(self,insertCommand):
        '''
        run an insert
        
        Args:
            insertCommand(string): the SPARQL INSERT command
          
        Returns:
            a response
        '''
        self.sparql.setRequestMethod(POSTDIRECTLY)
        response=None
        exception=None
        try:
            response=self.rawQuery(insertCommand, method=POST)
            #see https://github.com/RDFLib/sparqlwrapper/issues/159#issuecomment-674523696
            # dummy read the body
            response.response.read()
        except Exception as ex:
            exception=ex
            if self.debug:
                print (ex)
        return response,exception
    
    def getLocalName(self,name):
        '''
        retrieve valid localname from a string based primary key
        https://www.w3.org/TR/sparql11-query/#prefNames
        
        Args:
            name(string): the name to convert 
        
        Returns:
            string: a valid local name
        '''
        localName=''.join(ch for ch in name if ch.isalnum())
        return localName
    
    def insertListOfDicts(self,listOfDicts,entityType,primaryKey,prefixes,limit=None,batchSize=None, profile=False):
        '''
        insert the given list of dicts mapping datatypes 
        
        Args:
            entityType(string): the entityType to use as a 
            primaryKey(string): the name of the primary key attribute to use
            prefix(string): any PREFIX statements to be used
            limit(int): maximum number of records to insert
            batchSize(int): number of records to send per request
            
        Return:
            a list of errors which should be empty on full success
        
        datatype maping according to
        https://www.w3.org/TR/xmlschema-2/#built-in-datatypes
        
        mapped from 
        https://docs.python.org/3/library/stdtypes.html
        
        compare to
        https://www.w3.org/2001/sw/rdb2rdf/directGraph/
        http://www.bobdc.com/blog/json2rdf/
        https://www.w3.org/TR/json-ld11-api/#data-round-tripping
        https://stackoverflow.com/questions/29030231/json-to-rdf-xml-file-in-python
        '''
        if limit is not None:
            listOfDicts=listOfDicts[:limit]
        else:
            limit=len(listOfDicts)    
        total=len(listOfDicts)
        if batchSize is None:
            return self.insertListOfDictsBatch(listOfDicts, entityType, primaryKey, prefixes,total=total)
        else:
            startTime=time.time()
            errors=[]
            # store the list in batches
            for i in range(0, total, batchSize):
                recordBatch=listOfDicts[i:i+batchSize]
                batchErrors=self.insertListOfDictsBatch(recordBatch, entityType, primaryKey, prefixes,batchIndex=i,total=total,startTime=startTime)
                errors.extend(batchErrors)
            if self.profile:
                print("insertListOfDicts for %9d records in %6.1f secs" % (len(listOfDicts),time.time()-startTime),flush=True)
            return errors    
           
    def insertListOfDictsBatch(self,listOfDicts,entityType,primaryKey,prefixes,title='batch',batchIndex=None,total=None,startTime=None):                
        '''
        insert a Batch part of listOfDicts
        
        Args:
            entityType(string): the entityType to use as a 
            primaryKey(string): the name of the primary key attribute to use
            prefix(string): any PREFIX statements to be used
            title(string): the title to display for the profiling (if any)
            batchIndex(int): the start index of the current batch
            total(int): the total number of records for all batches
            starttime(datetime): the start of the batch processing
            
        Return:
            a list of errors which should be empty on full success
        '''
        errors=[]
        size=len(listOfDicts)
        if batchIndex is None:
            batchIndex=0    
        batchStartTime=time.time()
        if startTime is None:
            startTime=batchStartTime
        rdfprefix="PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>\n"
        insertCommand=f'{rdfprefix}{prefixes}\nINSERT DATA {{\n'
        for index,record in enumerate(listOfDicts):
            if not primaryKey in record:
                errors.append(f"missing primary key {primaryKey} in record {index}")
            else:    
                primaryValue=record[primaryKey]
                if primaryValue is None:
                    errors.append(f"primary key {primaryKey} value is None in record {index}")    
                else:    
                    encodedPrimaryValue=self.getLocalName(primaryValue)
                    tSubject=f"{entityType}__{encodedPrimaryValue}" 
                    insertCommand+=f'  {tSubject} rdf:type "{entityType}".\n' 
                    for keyValue in record.items():
                        key,value=keyValue
                        # convert key if necessary
                        key=self.getLocalName(key)
                        valueType=type(value)
                        if self.debug:
                            print("%s(%s)=%s" % (key,valueType,value))
                        tPredicate=f"{entityType}_{key}" 
                        tObject=value    
                        if valueType == str:
                            escapedString=self.controlEscape(value)
                            tObject='"%s"' % escapedString
                        elif valueType==int:
                            if self.typedLiterals:
                                tObject='"%d"^^<http://www.w3.org/2001/XMLSchema#integer>' %value
                            pass
                        elif valueType==float:
                            if self.typedLiterals:
                                tObject='"%s"^^<http://www.w3.org/2001/XMLSchema#decimal>' %value
                            pass
                        elif valueType==bool:
                            pass
                        elif valueType==datetime.date:
                            #if self.typedLiterals:
                            tObject='"%s"^^<http://www.w3.org/2001/XMLSchema#date>' %value
                            pass
                        elif valueType==datetime.datetime:
                            tObject='"%s"^^<http://www.w3.org/2001/XMLSchema#dateTime>' %value
                            pass
                        else:
                            errors.append("can't handle type %s in record %d" % (valueType,index))
                            tObject=None
                        if tObject is not None:   
                            insertRecord='  %s %s %s.\n' % (tSubject,tPredicate,tObject)
                            insertCommand+=insertRecord
        insertCommand+="\n}"
        if self.debug:
            print (insertCommand,flush=True)
        response,ex=self.insert(insertCommand)
        if response is None and ex is not None:
            errors.append("%s for record %d" % (str(ex),index))
        if self.profile:    
            print("%7s for %9d - %9d of %9d %s in %6.1f s -> %6.1f s" % (title,batchIndex+1,batchIndex+size,total,entityType,time.time()-batchStartTime,time.time()-startTime),flush=True)    
        return errors
    
    controlChars = [chr(c) for c in range(0x20)]
    
    @staticmethod
    def controlEscape(s):
        '''
        escape control characters
        
        see https://stackoverflow.com/a/9778992/1497139
        '''
        escaped=u''.join([c.encode('unicode_escape').decode('ascii') if c in SPARQL.controlChars else c for c in s])
        escaped=escaped.replace('"','\\"')
        return escaped

    def query(self,queryString,method=POST):
        '''
        get a list of results for the given query
        
        Args:
            queryString(string): the SPARQL query to execute
            method(string): the method eg. POST to use
            
        Returns:
            list: list of bindings
        '''
        queryResult=self.rawQuery(queryString,method=method) 
        if self.debug:
            print(queryString)
        if hasattr(queryResult, "info"):
            if "content-type" in queryResult.info():
                ct = queryResult.info()["content-type"] 
                if "text/html" in ct:
                    response=queryResult.response.read().decode()
                    if not "Success" in response:
                        raise("%s failed: %s", response)
                return None 
        jsonResult=queryResult.convert()
        return self.getResults(jsonResult)
    
    def queryAsListOfDicts(self,queryString,fixNone:bool=False,sampleCount:int=None):
        '''
        get a list of dicts for the given query (to allow round-trip results for insertListOfDicts)
        
        Args:
            queryString(string): the SPARQL query to execute
            fixNone(bool): if True add None values for empty columns in Dict
            sampleCount(int): the number of samples to check
            
        Returns:
            list: a list ofDicts
        '''
        records=self.query(queryString,method=self.method)
        listOfDicts=self.asListOfDicts(records,fixNone=fixNone,sampleCount=sampleCount)
        return listOfDicts
    
    @staticmethod
    def strToDatetime(value,debug=False):
        '''
        convert a string to a datetime
        Args:
            value(str): the value to convert
        Returns:
            datetime: the datetime
        '''
        dateFormat="%Y-%m-%d %H:%M:%S.%f"
        if "T" in value and "Z" in value:
            dateFormat="%Y-%m-%dT%H:%M:%SZ"
        dt=None
        try:
            dt=datetime.datetime.strptime(value,dateFormat)     
        except ValueError as ve:
            if debug:
                print(str(ve))
        return dt
    
    def asListOfDicts(self,records,fixNone:bool=False,sampleCount:int=None):
        '''
        convert SPARQL result back to python native
        
        Args:
            record(list): the list of bindings
            fixNone(bool): if True add None values for empty columns in Dict
            sampleCount(int): the number of samples to check
            
        Returns:
            list: a list of Dicts
        '''
        resultList=[]
        fields=None
        if fixNone:
            fields=LOD.getFields(records, sampleCount)
        for record in records:
            resultDict={}
            for keyValue in record.items():
                key,value=keyValue
                datatype=value.datatype
                if datatype is not None:
                    if datatype=="http://www.w3.org/2001/XMLSchema#integer":
                        resultValue=int(value.value) 
                    elif datatype=="http://www.w3.org/2001/XMLSchema#decimal":
                        resultValue=float(value.value)     
                    elif datatype=="http://www.w3.org/2001/XMLSchema#boolean":
                        resultValue=value.value in ['TRUE','true']    
                    elif datatype=="http://www.w3.org/2001/XMLSchema#date":
                        dt=datetime.datetime.strptime(value.value,"%Y-%m-%d")  
                        resultValue=dt.date()  
                    elif datatype=="http://www.w3.org/2001/XMLSchema#dateTime":
                        dt=SPARQL.strToDatetime(value.value,debug=self.debug) 
                        resultValue=dt
                    else:
                        # unsupported datatype
                        resultValue=value.value      
                else:
                    resultValue=value.value  
                resultDict[key]=resultValue
            if fixNone:
                for field in fields:
                    if not field in resultDict:
                        resultDict[field]=None
            resultList.append(resultDict)
        return resultList
    
    def printErrors(self,errors):
        '''
        print the given list of errors
        
        Args:
            errors(list): a list of error strings
            
        Returns:
            boolean: True if the list is empty else false
        '''
        if len(errors)>0:
            print("ERRORS:")
            for error in errors:
                print(error,flush=True,file=stderr)
            return True
        else:
            return False    